from pathlib import Path
from typing import Optional

from approvaltests.core import Comparator

from pytest_approvaltests_geo.geo_io import read_array_and_tags
from pytest_approvaltests_geo.scrubbers import TagsScrubber


class CompareGeoTiffs(Comparator):
    def __init__(self, tags_scrubber: Optional[TagsScrubber] = None):
        self._tags_scrubber = tags_scrubber

    def compare(self, received_path: str, approved_path: str) -> bool:
        received_path = Path(received_path)
        approved_path = Path(approved_path)
        if not received_path.exists() or not approved_path.exists():
            return False

        with read_array_and_tags(received_path) as (received_pixels, received_tags), \
                read_array_and_tags(approved_path) as (approved_pixels, approved_tags):

            if received_pixels.shape != approved_pixels.shape:
                return False

            if self._tags_scrubber:
                received_tags = self._tags_scrubber(received_tags)
                approved_tags = self._tags_scrubber(approved_tags)

            if received_tags != approved_tags:
                return False

            return received_pixels.equals(approved_pixels)



import os
from difflib import unified_diff
from pathlib import Path
from typing import Optional

import numpy as np
from approval_utilities.utils import to_json
from approvaltests import Reporter
from approvaltests.reporters import get_command_text
from xarray import DataArray

from pytest_approvaltests_geo.geo_io import read_array_and_tags
from pytest_approvaltests_geo.scrubbers import TagsScrubber


class ReportGeoTiffs(Reporter):
    def __init__(self, tags_scrubber: Optional[TagsScrubber] = None):
        self._tags_scrubber = tags_scrubber

    def report(self, received_path: str, approved_path: str) -> bool:
        received_path = Path(received_path)
        approved_path = Path(approved_path)

        if not approved_path.is_file():
            self._create_empty_geotiff(approved_path)

        with read_array_and_tags(received_path) as (received_pixels, received_tags), \
                read_array_and_tags(approved_path) as (approved_pixels, approved_tags):
            diff_tags = self._calculate_tags_diff(approved_path, approved_tags, received_path, received_tags)
            diff_pixels_msg = self._calculate_pixel_diff(approved_pixels, received_pixels)

            if diff_tags:
                diff_tags = f"Differences in meta data:\n{diff_tags}"
            if diff_pixels_msg:
                diff_pixels_msg = f"Differences in pixel data:\n{diff_pixels_msg}"
            if diff_tags or diff_pixels_msg:
                to_approve_msg = \
                    f"To approve run:\n {get_command_text(received_path.as_posix(), approved_path.as_posix())}"
                print('\n'.join([diff_tags, diff_pixels_msg, to_approve_msg]))
            return True

    def _calculate_tags_diff(self, approved_path, approved_tags, received_path, received_tags):
        if self._tags_scrubber:
            approved_text = f"{self._tags_scrubber(approved_tags)}"
            received_text = f"{self._tags_scrubber(received_tags)}"
        else:
            approved_text = f"{to_json(approved_tags)}\n"
            received_text = f"{to_json(received_tags)}\n"
        return "\n".join(unified_diff(
            approved_text.splitlines(),
            received_text.splitlines(),
            os.path.basename(approved_path),
            os.path.basename(received_path)
        )).strip()

    @staticmethod
    def _calculate_pixel_diff(approved_pixels, received_pixels):
        diff_pixels = np.abs(received_pixels - approved_pixels)
        diff_min = np.nanmin(diff_pixels)
        diff_max = np.nanmax(diff_pixels)
        diff_mean = np.nanmean(diff_pixels)
        diff_median = np.nanmedian(diff_pixels)
        if diff_min != 0 or diff_max != 0:
            return f"pixel differences statistics:\n" \
                   f"min={diff_min}, max={diff_max}, mean={diff_mean}, median={diff_median}"
        return ""

    @staticmethod
    def _create_empty_geotiff(path: Path):
        DataArray([[0]], dims=['y', 'x']).rio.to_raster(path)



from typing import Dict, Tuple, Callable

from approvaltests import Options, ScenarioNamer, Namer
from approvaltests.namer import NamerBase

from pytest_approvaltests_geo.scrubbers import TagsScrubber


class GeoOptions(Options):
    _TAGS_SCRUBBER_FUNC = "tags_scrubber_func"
    _NAMER_WRAPPER_SCENARIO_BY_TAGS = "NamerWrapperScenarioByTags"

    @classmethod
    def from_options(cls, options: Options) -> "GeoOptions":
        return GeoOptions(options.fields)

    def scrub_tags(self, data: Dict):
        if self.has_tags_scrubber():
            return self.fields[GeoOptions._TAGS_SCRUBBER_FUNC](data)
        return data

    def with_tags_scrubber(self, scrubber_func: TagsScrubber) -> "GeoOptions":
        return GeoOptions({**self.fields, **{GeoOptions._TAGS_SCRUBBER_FUNC: scrubber_func}})

    def has_tags_scrubber(self):
        return GeoOptions._TAGS_SCRUBBER_FUNC in self.fields

    def with_scenario_by_tags(self, tags_to_names: Callable[[Dict], Tuple[str, ...]]):
        return GeoOptions({**self.fields, **{GeoOptions._NAMER_WRAPPER_SCENARIO_BY_TAGS: tags_to_names}})

    def has_scenario_by_tags(self):
        return GeoOptions._NAMER_WRAPPER_SCENARIO_BY_TAGS in self.fields

    def wrap_namer_in_tags_scenario(self, namer: NamerBase, tags: Dict):
        if self.has_scenario_by_tags():
            return ScenarioNamer(namer, *self.fields[GeoOptions._NAMER_WRAPPER_SCENARIO_BY_TAGS](tags))

    @property
    def namer(self) -> Namer:
        return self.fields.get("namer")
